from django.db.models import Q

from aio_client.provider.api import PostProviderRequest
from aio_client.provider.api import push_request
from m3.plugins import ExtensionManager

from kinder.core.declaration.models import Declaration
from kinder.core.dict.models import DULTypeEnumerate
from kinder.core.helpers import get_q

from concentrator.smev3_v321.base import constants as base_enums
from concentrator.smev3_v321.base.utils import BaseExecutor
from concentrator.smev3_v321.base.utils import SMEV3Response
from concentrator.smev3_v321.base.utils import render_type2xml
from concentrator.smev3_v321.order.helpers import (
    OrderRequestSMEV3RequestManager)
from concentrator.smev3_v321.order.request_context import OrderRequestContext
from concentrator.smev3_v321.service_types import kinder_conc
from kinder.core.declaration_status.enum import DSS

from .constants import APPLICATION_ORDER_INFO_REQUEST
from .constants import APPLICATION_ORDER_INFO_REQUEST_SUCCESS_CODE
from .constants import APPLICATION_ORDER_INFO_REQUEST_SUCCESS_COMMENT
from .constants import APPLICATION_ORDER_INFO_REQUEST_ERROR_CODE
from .constants import APPLICATION_ORDER_INFO_REQUEST_ERROR_COMMENT
from .utils import update_declaration_from_request


class ApplicationOrderInfoRequestExecutor(BaseExecutor):
    """Исполнитель сервиса ApplicationOrderInfoRequest."""

    name_service = APPLICATION_ORDER_INFO_REQUEST
    type_service = kinder_conc.ApplicationOrderInfoRequestType

    @staticmethod
    def create_response(order_id, tech_code, comment, message):
        response_body = render_type2xml(
            kinder_conc.FormDataResponseType(
                changeOrderInfo=kinder_conc.changeOrderInfoType(
                    orderId=kinder_conc.orderIdType(order_id),
                    statusCode=kinder_conc.statusCodeType(
                        techCode=tech_code),
                    comment=comment)),
            name_type='FormDataResponse')

        return push_request(PostProviderRequest(
            origin_message_id=message['origin_message_id'],
            body=response_body,
            message_type=message['message_type'],
            replay_to=message['replay_to'],
            is_test_message=message['is_test_message'])), response_body

    @classmethod
    def _process(cls, request_body):
        """Доп.метод для удобства обработки и возвращения результата."""

        request_data = request_body.ApplicationOrderInfoRequest
        order_id = request_data.orderId
        child_info = request_data.ChildInfo
        child_birth_doc_rf = child_info.ChildBirthDocRF

        if child_birth_doc_rf:
            child_doc_info = child_birth_doc_rf
            dul_filter = (
                Q(children__dul_type=DULTypeEnumerate.SVID)
                & Q(children__dul_series=child_doc_info.ChildBirthDocSeries)
            )
        else:
            child_doc_info = child_info.ChildBirthDocForeign
            dul_filter = Q(children__dul_type=DULTypeEnumerate.INT_SVID)
            series = child_doc_info.ChildBirthDocSeries
            # Серия может быть не заполнена у иностранных документов
            dul_filter &= (
                Q(children__dul_series=series) if series else
                get_q('children__dul_series', is_empty=True)
            )
        dul_filter &= Q(children__dul_number=child_doc_info.ChildBirthDocNumber)

        # Фильтр для поиска заявления по ФИ ребенка, Дате рождения
        # и ДУЛ (серия, номер).
        base_filter = (
            Q(children__surname__iexact=child_info.ChildSurname)
            & Q(children__firstname__iexact=child_info.ChildName)
            & Q(children__date_of_birth=child_info.ChildBirthDate)
            & dul_filter
        )

        declarations_query = Declaration.objects.filter(base_filter).exclude(
            status__code=DSS.ARCHIVE
        ).select_related('children')

        # Проверка наличия заявления
        if not declarations_query:
            tech_code = APPLICATION_ORDER_INFO_REQUEST_ERROR_CODE
            comment = APPLICATION_ORDER_INFO_REQUEST_ERROR_COMMENT
            return tech_code, comment

        has_full_declaration, error_message = False, ''

        for declaration in declarations_query:
            # Проверка наличия нужных данных у представителя и попытка
            # обновления их данными из запроса (данные могут измениться).
            update_info = update_declaration_from_request(
                declaration, request_body.ApplicationOrderInfoRequest)
            has_full_info, delegate_id, error_message = update_info

            if not has_full_info:
                continue
            has_full_declaration = True

            if child_birth_doc_rf:
                zags_act_number = child_birth_doc_rf.ChildBirthDocActNumber
                ExtensionManager().execute(
                    'kinder.plugins.tatarstan_specifics.extensions'
                    '.update_child_zags_act_number',
                    declaration, zags_act_number)

            OrderRequestSMEV3RequestManager(
                OrderRequestContext(**{
                    'declaration_id': declaration.id,
                    'order_id': order_id,
                    'delegate_id': delegate_id,
                })
            ).apply_async()

        # Если нет ни одного заявлений с полными данными, прекращаем обработку
        if not has_full_declaration:
            return APPLICATION_ORDER_INFO_REQUEST_ERROR_CODE, error_message

        tech_code = APPLICATION_ORDER_INFO_REQUEST_SUCCESS_CODE
        comment = APPLICATION_ORDER_INFO_REQUEST_SUCCESS_COMMENT

        return tech_code, comment

    @classmethod
    def process(cls, message, request_body, **kwargs):
        request_data = request_body.ApplicationOrderInfoRequest
        tech_code, comment = cls._process(request_body)
        response, response_body = cls.create_response(
            request_data.orderId, tech_code, comment, message)

        return SMEV3Response(
            response,
            {
                'method_name': cls.name_service,
                'response': response_body,
                'result': base_enums.SUCCESS_MESSAGE
            })
